#if !defined(DOMPARSER_HPP)
#define DOMPARSER_HPP

#include <assert.h>

#include <iostream>
#include <vector>
#include <string>
#include <map>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>

#include <xercesc/util/PlatformUtils.hpp>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMWriter.hpp>

#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>

#include "ko/DOMTreeErrorReporter.hpp"
#include "ko/DOMPrintFilter.hpp"
#include "ko/DOMPrintErrorHandler.hpp"
#include "ko/OStreamFormatTarget.hpp"


class DOMTreeErrorReporter;

XERCES_CPP_NAMESPACE_BEGIN

template<class S>
class DOMParser
{
public:

  bool gDoNamespaces;
  bool gDoSchema;
  bool gSchemaFullChecking;
  bool gDoCreate;
  XMLCh* gEncodingName;

  // options for DOMWriter's features
  const XMLCh* gOutputEncoding;
  const XMLCh* gMyEOLSequence;
  
  bool gSplitCdataSections;
  bool gDiscardDefaultContent;
  bool gUseFilter;
  bool gFormatPrettyPrint;
  
  XercesDOMParser::ValSchemes gValScheme;
  
  std::string rootTag;

  XercesDOMParser *parser;
  DOMTreeErrorReporter *errReporter;

  DOMParser() :
    gDoNamespaces(true), gDoSchema(true),
    gSchemaFullChecking(true), gDoCreate(true),
    gEncodingName(0), gValScheme(XercesDOMParser::Val_Auto),
    gOutputEncoding(0), gMyEOLSequence(0),
    gSplitCdataSections(true), gDiscardDefaultContent(true),
    gUseFilter(false), gFormatPrettyPrint(true)
  {
    //gDoSchema = false;
    //gSchemaFullChecking = false;
    
    //std::cerr << "xml init" << std::endl;
    // Initialize the XML4C2 system
    try
    {
      XMLPlatformUtils::Initialize();
    }
    
    catch(const XMLException& toCatch)
    {
      std::cerr << "Error during Xerces-c Initialization.\n"
	   << "  Exception message:"
	   << toCatch.getMessage() << std::endl;
    }
    //std::cerr << "xml init done" << std::endl;
    
    errReporter = new DOMTreeErrorReporter();
    parser = new XercesDOMParser;
    
    parser->setErrorHandler(errReporter);
    //parser->setToCreateXMLDeclTypeNode(true);
    parser->setValidationScheme(gValScheme);
    parser->setDoNamespaces(gDoNamespaces);
    parser->setDoSchema(gDoSchema);
    parser->setValidationSchemaFullChecking(gSchemaFullChecking);
    
  }

  void parse(const char *gXmlFile, S &target)
  {
    //
    //  Parse the XML file, catching any XML exceptions that might propogate
    //  out of it.
    //
    bool errorsOccured = false;
    try
    {
      parser->parse(gXmlFile);
      int errorCount = parser->getErrorCount();
      if (errorCount > 0)
	errorsOccured = true;
    }
    
    catch (const XMLException& e)
    {
      std::cerr << "An error occured during parsing\n   Message: "
	   << e.getMessage() << std::endl;
      errorsOccured = true;
    }
    
    catch (const DOMException& e)
    {
      std::cerr << "A DOM error occured during parsing\n   DOMException code: "
	   << e.code << std::endl;
      errorsOccured = true;
    }
    
    catch (...)
    {
      std::cerr << "An error occured during parsing\n " << std::endl;
      errorsOccured = true;
    }
    
    // If the parse was successful, output the document data from the DOM tree
    if (!errorsOccured && !errReporter->getSawErrors())
    {
      //DOMNode root = parser->getDocument();
      //DOMDocument doc = (DOMDocument &)root;

      DOMNode *root = parser->getDocument();
      DOMDocument &doc = (DOMDocument &)*root;
      
      XMLCh *tmpstr = XMLString::transcode(rootTag.c_str());

      DOMNodeList *childlist =
	doc.getElementsByTagName(tmpstr);
      XMLString::release(&tmpstr);
      if(childlist->getLength() != 0)
      {
	assert(childlist->item(0)->getNodeType() ==
	       DOMNode::ELEMENT_NODE);  
	
	DOMElement *child = (DOMElement *)childlist->item(0);
	
	std::cerr << "parsing" << std::endl;

	translate(*child, target);
      }
    }
    else
      std::cerr << "parse error" << std::endl;
    
    delete errReporter;
    delete parser;
    //XMLPlatformUtils::Terminate();
  }
  
  //NB! Memory leak detected, possibly inside Xerces
  void write(std::ostream &os, S &source)
  {
    DOMDocument *doc = create(source);

    DOMPrintFilter *myFilter = 0;

    try
    {
      // get a serializer, an instance of DOMWriter
      XMLCh tempStr[100];
      XMLString::transcode("LS", tempStr, 99);
      DOMImplementation *impl =
	DOMImplementationRegistry::getDOMImplementation(tempStr);
      DOMWriter *theSerializer =
	((DOMImplementationLS*)impl)->createDOMWriter();

      // set user specified end of line sequence and output encoding
      theSerializer->setNewLine(gMyEOLSequence);
      theSerializer->setEncoding(gOutputEncoding);

      // plug in user's own filter
      if (gUseFilter)
      {
	// even we say to show attribute, but the DOMWriter
	// will not show attribute nodes to the filter as
	// the specs explicitly says that DOMWriter shall
	// NOT show attributes to DOMWriterFilter.
	//
	// so DOMNodeFilter::SHOW_ATTRIBUTE has no effect.
	// same DOMNodeFilter::SHOW_DOCUMENT_TYPE, no effect.
	//
	myFilter =
	  new DOMPrintFilter(DOMNodeFilter::SHOW_ELEMENT   |
			     DOMNodeFilter::SHOW_ATTRIBUTE |
			     DOMNodeFilter::SHOW_DOCUMENT_TYPE
			     );
	theSerializer->setFilter(myFilter);
      }

      // plug in user's own error handler
      DOMErrorHandler *myErrorHandler = new DOMPrintErrorHandler();
      theSerializer->setErrorHandler(myErrorHandler);

      // set feature if the serializer supports the feature/mode
      if(theSerializer->canSetFeature(XMLUni::fgDOMWRTSplitCdataSections,
				       gSplitCdataSections))
	theSerializer->setFeature(XMLUni::fgDOMWRTSplitCdataSections,
				  gSplitCdataSections);
			
      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent,
				       gDiscardDefaultContent))
	theSerializer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent,
				  gDiscardDefaultContent);

      if (theSerializer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint,
				       gFormatPrettyPrint))
	theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint,
				  gFormatPrettyPrint);

      //
      // Plug in a format target to receive the resultant
      // XML stream from the serializer.
      //
      // StdOutFormatTarget prints the resultant XML stream
      // to stdout once it receives any thing from the serializer.
      //
      XMLFormatTarget *myFormTarget =
	new OStreamFormatTarget(os);

      // get the DOM representation
      //DOMNode *doc = parser->getDocument();

      //
      // do the serialization through DOMWriter::writeNode();
      //
      theSerializer->writeNode(myFormTarget, *doc);

      delete theSerializer;

      //
      // Filter, formatTarget and error handler
      // are NOT owned by the serializer.
      //
      delete myFormTarget;
      delete myErrorHandler;

      if (gUseFilter)
	delete myFilter;      	

    }
    catch (XMLException& e)
    {
      std::cerr << "An error occurred during creation of output transcoder. Msg is:"
	   << std::endl
	   << StrX(e.getMessage()) << std::endl;
      //retval = 4;
    }


    delete doc;
    //writeDocument(os, *doc);

    //XMLPlatformUtils::Terminate();

  }
  
  virtual void translate(const DOMElement &node, S &target) = 0;

  virtual DOMDocument *create(S &source) = 0;

};

XERCES_CPP_NAMESPACE_END

#endif // DOMParser

