#include <iostream>
#include <strstream>
#include <stdio.h>

#include "ko/PhysicalDolfinInterface.hpp"

namespace ko
{

PhysicalDolfinInterface::PhysicalDolfinInterface(dolfin::Mesh& mesh,
						 Physical& physical,
						 dolfin::Function& velocity,
						 dolfin::real young,
						 dolfin::real damp) :
  mesh(mesh), boundary(mesh), physical(physical), velocity(velocity)

{
  build();

  m.young = 1.0 * young;
  m.damp = 0.05 * damp;
  m.friction = 0.4;
  m.density = 1e3;
  m.strength = 1e10;
}

void PhysicalDolfinInterface::updatePhysical()
{
  for(dolfin::NodeIterator nmesh(boundary); !nmesh.end(); ++nmesh)
  {
    dolfin::Node& node = *nmesh;
    
    updateNode(node);
  }
}

void PhysicalDolfinInterface::build()
{
  elements.resize(mesh.noNodes());
  nodes.resize(mesh.noNodes());

  //for(dolfin::NodeIterator nmesh(boundary); !nmesh.end(); ++nmesh)
  for(dolfin::NodeIterator nmesh(boundary); !nmesh.end(); ++nmesh)
  {
    dolfin::Node& node = *nmesh;

    addNode(node);
  }

  for(dolfin::FaceIterator c(boundary); !c.end(); ++c)
  {
    dolfin::Face &cell = *c;
    
    addBoundaryCell(cell);
  }

//   for(dolfin::CellIterator c(mesh); !c.end(); ++c)
//   {
//     dolfin::Cell &cell = *c;
    
//     addBoundaryCell(cell);
//   }



  m.name = "material";

  physical.materials.push_back(&m);




//   Connect nearby elements together (a tetrahedral pattern is too sparse)
  for(std::list<Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element &e = **it;
    
    for(std::list<Connection *>::iterator it2 = e.connections.begin();
	it2 != e.connections.end(); it2++)
    {
      Connection &ce = **it2;
      
      Element *etmp = 0;
      
      if(ce.element1 == &e)
      {
	etmp = ce.element2;
      }
      else
      {
	etmp = ce.element1;
      }
      
      for(std::list<Connection *>::iterator it3 = etmp->connections.begin();
	  it3 != etmp->connections.end(); it3++)
      {
	Connection &ce = **it3;
	
	Element *e2 = 0;
	
	if(ce.element1 == &e)
	{
	  e2 = ce.element2;
	}
	else
	{
	  e2 = ce.element1;
	}
	
	Vector r(3);
	r = e.position;
	r.axpy(-1, e2->position);
	
	dolfin::real distance;
	
	distance = r.norm();
	
	if(distance <= 3.0 * e.radius)
	{
	  if(!e.isConnected(*e2))
	  {
	    Connection &c = *(new Connection());
	    
	    c.connect(e, *e2);
	    c.length = distance;
	    physical.addConnection(c);
	    
// 	      std::cerr << "added connection" << std::endl;
	  }
	}
      }
    }
  }
}

void PhysicalDolfinInterface::addNode(dolfin::Node& node)
{
  Element &e = *(new Element());

  physical.elements.push_back(&e);
  elements[node.id()] = &e;
  nodes[node.id()] = &node;

  updateNode(node);

  e.material = &m;

  e.index = node.id();

  std::ostrstream name;
  name << node.id() << std::ends;
  
  e.name = name.str();

  // Determine radius
  // 0.3 * max of neighbor cell diameter

//   dolfin::real maxr = DBL_MAX;
  dolfin::real sum = 0.0;
  for(dolfin::CellIterator cell(node); !cell.end(); ++cell)
  {
    dolfin::Cell &c = *cell;
//     maxr = std::min(maxr, c.diameter());
    dolfin::cout << "h: " << c.diameter() << dolfin::endl;

    sum += c.diameter();
  }
  sum /= node.noCellNeighbors();
//   maxr *= 0.3;
  sum *= 0.40;

  e.radius = sum;
}

void PhysicalDolfinInterface::addCell(dolfin::Cell &cell)
{
  for(int i = 0; i < 4; i++)
  {
    for(int j = 0; j < 4; j++)
    {
      Element &e0 = *(elements[cell.node(i).id()]);
      Element &e1 = *(elements[cell.node(j).id()]);

      if((i != j) && !e0.isConnected(e1))
      {
	Connection &c = *(new Connection());
	
	Vector r(3);
	double rnorm;
	
	r = e0.position;
	r.axpy(-1, e1.position);
	
	rnorm = r.norm();
	
	c.connect(e0, e1);
	c.length = rnorm;
	
	physical.connections.push_back(&c);
      }
    }
  }
}

void PhysicalDolfinInterface::addBoundaryCell(dolfin::Face &cell)
{
  for(dolfin::EdgeIterator nedge(cell); !nedge.end(); ++nedge)
  {
    dolfin::Edge& edge = *nedge;

    Element &e0 = *(elements[edge.node(0).id()]);
    Element &e1 = *(elements[edge.node(1).id()]);

    if(!e0.isConnected(e1))
    {
      Connection &c = *(new Connection());
      
      Vector r(3);
      double rnorm;
      
      r = e0.position;
      r.axpy(-1, e1.position);
      
      rnorm = r.norm();
      
      c.connect(e0, e1);
      c.length = rnorm;
      
      physical.connections.push_back(&c);
    }
  }

//   for(int i = 0; i < 4; i++)
//   {
//     for(int j = 0; j < 4; j++)
//     {
//       Element &e0 = *(elements[cell.node(i).id()]);
//       Element &e1 = *(elements[cell.node(j).id()]);

//       if((i != j) && !e0.isConnected(e1))
//       {
// 	Connection &c = *(new Connection());
	
// 	Vector r(3);
// 	double rnorm;
	
// 	r = e0.position;
// 	r.axpy(-1, e1.position);
	
// 	rnorm = r.norm();
	
// 	c.connect(e0, e1);
// 	c.length = rnorm;
	
// 	physical.connections.push_back(&c);
//       }
//     }
//   }
}

void PhysicalDolfinInterface::updateNode(dolfin::Node& node)
{
  Element &e = *(elements[node.id()]);

  e.position(0) = node.coord().x;
  e.position(1) = node.coord().y;
  e.position(2) = node.coord().z;

  e.velocity(0) = velocity(node, 0);
  e.velocity(1) = velocity(node, 1);
  e.velocity(2) = velocity(node, 2);
}

}
