#include <stdio.h>
#include <float.h>
#include <iostream>
#include <iomanip>
#include <map>
#include <sys/time.h>
#include "ko/BasicEulerIntegrator.hpp"

using namespace std;

namespace ko
{

BasicEulerIntegrator::BasicEulerIntegrator(Physical &physical) :
  NumericalIntegrator(physical), method(FORWARD_EULER), order(DG0),
  demandConvergence(true), eulerTolerance(1E-8), eulerMaxIters(100)
{
}

bool BasicEulerIntegrator::integrate(double t0, double &t1)
{
  double stepsize = fixedk;
  t1 = t0 + fixedk;
  int i = 0;

  bool converged = false;

  Vector zero(3);

  for(list<Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element &e = **it;
      
    e.position0 = e.position;
    e.velocity0 = e.velocity;
    e.force0 = e.force;
  }

  dynamicElements.clear();
  for(list<Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element &e = **it;
      
    if(!e.artificial)
    {
      dynamicElements.push_back(&e);
      e.index = i;

      i++;
    }
  }

  n = dynamicElements.size();

  if(order == CG1)
  {
    cG1u10.init(6 * n);

    residualFunctiondG0(cG1u10, stepsize);

    for(list<Element *>::iterator it = dynamicElements.begin();
	it != dynamicElements.end(); it++)
    {
      Element &e = **it;
      
      int i = e.index;

      cG1u10(3 * i + 0) += e.position(0);
      cG1u10(3 * i + 1) += e.position(1);
      cG1u10(3 * i + 2) += e.position(2);

      cG1u10(3 * i + 3 * n + 0) += e.velocity(0);
      cG1u10(3 * i + 3 * n + 1) += e.velocity(1);
      cG1u10(3 * i + 3 * n + 2) += e.velocity(2);
    }

  }


  //cerr << "n: " << n << endl;

  if(method == BACKWARD_EULER)
  {
    converged = newton(stepsize);
  }
  else if(method == FORWARD_EULER)
  {
    converged = fixedPoint(stepsize);
  }
  else
  {
    cerr << "unknown integration method" << endl;
  }

  if(!converged)
  {
    cerr << "numerical integration produced no correct solution" << endl;
    
    if(demandConvergence)
    {
      for(list<Element *>::iterator it = physical.elements.begin();
	  it != physical.elements.end(); it++)
      {
	Element &e = **it;
	
	e.position = e.position0;
	e.velocity = e.velocity0;
      }


      return false;
    }
    else
    {
      cerr << "continuing anyway" << endl;
      return true;
    }
  }
  return true;
}


bool BasicEulerIntegrator::fixedPoint(double stepsize)
{
  Vector zero(3);
  LargeVector residual(6 * n);

  bool converged = false;
  int iter = 1;
  double rdprev, rd;

  while(true)
  {
    residual = 0;
    residualFunction(residual, stepsize);

    // Add the residual to compute the next iterate
    for(list<Element *>::iterator it = dynamicElements.begin();
	it != dynamicElements.end(); it++)
    {
      Element &e = **it;

      int i = e.index;

      for(int j = 0; j < 3; j++)
      {
	e.position(j) += residual(3 * i + j);
	e.velocity(j) += residual(3 * i + 3 * n + j);
      }
    }      

    rd = residual.norm(dolfin::Vector::linf);

    cerr << "rd.norm(linf): " << rd << endl;

    if(rd <= eulerTolerance && iter >= 1)
    {
      converged = true;
      break;
    }
    else if(iter > eulerMaxIters || (rd > rdprev && iter > 4))
    {
      converged = false;
      break;
    }

    iter++;
    rdprev = rd;
  }



//   while(true)
//   {
//     residualFunction(residual, stepsize);

//     rd = residual.norm(dolfin::Vector::linf);

//     cerr << "rd.norm(0): " << rd << endl;
//     if(rd <= eulerTolerance && iter > 1)
//     {
//       converged = true;
//       break;
//     }
//     else if(iter > eulerMaxIters || (rd > rdprev && iter > 1))
//     {
//       converged = false;
//       break;
//     }
//     else
//     {
//       // We already have the residual, so just add it

//       for(list<Element *>::iterator it = dynamicElements.begin();
// 	  it != dynamicElements.end(); it++)
//       {
// 	Element &e = **it;

// 	int i = e.index;

// 	for(int j = 0; j < 3; j++)
// 	{
// 	  e.position(j) += residual(3 * i + j);
// 	  e.velocity(j) += residual(3 * i + 3 * n + j);
// 	}
//       }      
//     }

//     iter++;
//     rdprev = rd;
//   }



  return converged;
}


void BasicEulerIntegrator::residualFunction(LargeVector &residual,
				    double stepsize)
{
  if(order == CG1)
  {
    for(list<Element *>::iterator it = dynamicElements.begin();
	it != dynamicElements.end(); it++)
    {
      Element *e = *it;

      residualFunctioncG1(residual, stepsize, e);
    }
  }
  else
  {
    residualFunctiondG0(residual, stepsize);
  }

//   for(list<Element *>::iterator it = dynamicElements.begin();
//       it != dynamicElements.end(); it++)
//   {
//     Element *e = *it;
    
//     if(order == CG1)
//     {
//       residualFunctioncG1(residual, stepsize, e);
//     }
//     else
//     {
//       residualFunctiondG0(residual, stepsize, e);
//     }
//   }
}

void BasicEulerIntegrator::residualFunction(LargeVector &residual,
				    double stepsize,
				    Element *e)
{
  if(order == CG1)
  {
    residualFunctioncG1(residual, stepsize, e);
  }
  else
  {
    residualFunctiondG0(residual, stepsize, e);
  }
}

void BasicEulerIntegrator::residualFunctiondG0(LargeVector &residual,
				       double stepsize)
{
  for(list<Element *>::iterator it = dynamicElements.begin();
      it != dynamicElements.end(); it++)
  {
    Element *e = *it;

    e->force = 0;
  }

  for(list<Connection *>::iterator it = physical.connections.begin();
      it != physical.connections.end(); it++)
  {
    Connection &c = **it;

    static Vector force(3);

    Element *e1 = c.element1;
    Element *e2 = c.element2;

    mmodel.force(c, *e1, force);
    e1->force += force;
    force *= -1;
    e2->force += force;

    // Project force on connection
    static Vector r(3), tmp(3);
    double rnorm;

    r = e1->position;
    r.axpy(-1, e2->position);

    rnorm = r.norm();

    tmp = r;
    tmp *= (force * r) / (rnorm * rnorm);

    c.tension = tmp.norm();

    if(force * r < 0)
      c.tension *= -1;

//     cerr << "tension: " << c.tension << endl;
  }

  for(list<Element *>::iterator it = dynamicElements.begin();
      it != dynamicElements.end(); it++)
  {
    Element *e = *it;
    
    int i = e->index;
    
    static Vector force(3), acceleration(3);
    
    double mass = e->material->density * pow(2 * e->radius, 3);
    

    // FIXME: Compute local element force manually for now

    // mmodel.force(*e, force);
    force = 0.0;
    force.axpy(mass, mmodel.g);
    e->force += force;

    
    acceleration = e->force;
    acceleration *= 1 / mass;
    
    // Top half of residual is p' = v and lower half is v' = F / m
    
    // Backward Euler
    dolfin::real block[6];
    int dofs[6];

    for(int j = 0; j < 3; j++)
    {
//       residual(3 * i + j) = e->position0(j) - e->position(j) +
// 	stepsize * e->velocity(j);
//       residual(3 * i + 3 * n + j) = e->velocity0(j) - e->velocity(j) +
// 	stepsize * acceleration(j);

      dofs[j] = 3 * i + j;
      block[j] = e->position0(j) - e->position(j) +
 	stepsize * e->velocity(j);

      dofs[j + 3] = 3 * i + 3 * n + j;
      block[j + 3] = e->velocity0(j) - e->velocity(j) +
 	stepsize * acceleration(j);
    }
    residual.add(block, dofs, 6);
      
  }
  residual.apply();

//   for(list<Element *>::iterator it = dynamicElements.begin();
//       it != dynamicElements.end(); it++)
//   {
//     Element *e = *it;
    
//     residualFunctiondG0(residual, stepsize, e);
//   }
}

void BasicEulerIntegrator::residualFunctioncG1(LargeVector &residual,
				       double stepsize)
{
  for(list<Element *>::iterator it = dynamicElements.begin();
      it != dynamicElements.end(); it++)
  {
    Element *e = *it;
    
    residualFunctioncG1(residual, stepsize, e);
  }
}

void BasicEulerIntegrator::residualFunctiondG0(LargeVector &residual,
				       double stepsize,
				       Element *e)
{
  int i = e->index;

  static Vector force(3), acceleration(3);

  mmodel.force(*e, force);

  e->force = force;

  double mass = e->material->density * pow(2 * e->radius, 3);
  
  acceleration = e->force;
  acceleration *= 1 / mass;
  
  // Top half of residual is p' = v and lower half is v' = F / m
  
  // Backward Euler

  for(int j = 0; j < 3; j++)
  {
    residual(3 * i + j) = e->position0(j) - e->position(j) +
      stepsize * e->velocity(j);
    residual(3 * i + 3 * n + j) = e->velocity0(j) - e->velocity(j) +
      stepsize * acceleration(j);
  }
}

void BasicEulerIntegrator::residualFunctioncG1(LargeVector &residual,
				       double stepsize,
				       Element *e)
{
  int i = e->index;

  residualFunctiondG0(residual, stepsize, e);

  residual(3 * i + 0) += cG1u10(3 * i + 0) + e->position(0);
  residual(3 * i + 0) *= 0.5;
  residual(3 * i + 0) -= e->position(0);
  residual(3 * i + 1) += cG1u10(3 * i + 1) + e->position(1);
  residual(3 * i + 1) *= 0.5;
  residual(3 * i + 1) -= e->position(1);
  residual(3 * i + 2) += cG1u10(3 * i + 2) + e->position(2);
  residual(3 * i + 2) *= 0.5;
  residual(3 * i + 2) -= e->position(2);
  
  residual(3 * i + 3 * n + 0) += cG1u10(3 * i + 3 * n + 0) +
    e->velocity(0);
  residual(3 * i + 3 * n + 0) *= 0.5;
  residual(3 * i + 3 * n + 0) -= e->velocity(0);
  residual(3 * i + 3 * n + 1) += cG1u10(3 * i + 3 * n + 1) +
    e->velocity(1);
  residual(3 * i + 3 * n + 1) *= 0.5;
  residual(3 * i + 3 * n + 1) -= e->velocity(1);
  residual(3 * i + 3 * n + 2) += cG1u10(3 * i + 3 * n + 2) +
    e->velocity(2);
  residual(3 * i + 3 * n + 2) *= 0.5;
  residual(3 * i + 3 * n + 2) -= e->velocity(2);
  
}

bool BasicEulerIntegrator::newton(double stepsize)
{
  bool converged = false;

  map<Element *, int> lookup;

  int iters = 0, j;

  Vector eresidual(3);
  Vector zero(3);

  //Matrix J;
  struct timeval before, after;


  if(eulerTolerance <= DBL_EPSILON)
    eulerTolerance = sqrt(DBL_EPSILON);


  int count = 0;
  for(list<Element *>::iterator it = dynamicElements.begin();
      it != dynamicElements.end(); it++)
  {
    Element *e = *it;

    lookup[e] = count;

    count++;
  }


  double hd, pnorm;

  LargeVector residual(6 * n), sk(6 * n),
    rd(6 * n);

  //residual.resize(dynamicElements.size() * 3);
  //sk.resize(dynamicElements.size() * 3);
  //pi.resize(dynamicElements.size() * 3);
  //piprev.resize(dynamicElements.size() * 3);


  Matrix J(6 * n, 6 * n);

  double elapsed;

  for(int i = 0; i < eulerMaxIters; i++)
  {
    iters = i + 1;

    residualFunction(residual, stepsize);
    residual *= -1;

    //gettimeofday(&before, NULL);
	
    jacobian(lookup, J, stepsize);

    //J.show();
    
    //gettimeofday(&after, NULL);
	
    //elapsed = (1E6 * after.tv_sec + after.tv_usec) -
    //(1E6 * before.tv_sec + before.tv_usec);
	
    //cerr << "jacobian: " << elapsed << endl;
    
    //gettimeofday(&before, NULL);

    if(CGSolve(J, sk, residual, 0, 0) != 0)
    {
      cerr << "Krylov problem" << endl;
      break;
    }
	
    //gettimeofday(&after, NULL);
	
    //elapsed = (1E6 * after.tv_sec + after.tv_usec) -
    //(1E6 * before.tv_sec + before.tv_usec);
	
    //cerr << "solve: " << elapsed << endl;

    j = 0;
    for(list<Element *>::iterator it = dynamicElements.begin();
	it != dynamicElements.end(); it++)
    {
      Element &e = **it;
	    
      e.position(0) += sk(3 * j + 0);
      e.position(1) += sk(3 * j + 1);
      e.position(2) += sk(3 * j + 2);

      e.velocity(0) += sk(3 * j + 3 * n + 0);
      e.velocity(1) += sk(3 * j + 3 * n + 1);
      e.velocity(2) += sk(3 * j + 3 * n + 2);

      j++;
    }
	
    residualFunction(rd, stepsize);

    cerr << "rd: " << rd.norm(dolfin::Vector::linf) << endl;
    
    cerr << "change: " << sk.norm() << endl;

    if(rd.norm(dolfin::Vector::linf) <= eulerTolerance)
    {
      converged = true;
      break;
    }
    else
    {
      // Check divergence?
    }
	
    fprintf(stderr, "iter: %d\n", iters);
  }
  fprintf(stderr, "iters: %d\n", iters);
    
  return converged;
}

void BasicEulerIntegrator::jacobian(map<Element *, int> lookup, Matrix &J,
				     double stepsize)
{
  cerr << "computing Jacobian " << endl;

  Matrix Processed(n, n);
  LargeVector residual0(6 * n), residual1(6 * n);

  Vector zero(3);

  int nderivs = 0, ej, ei, i, j, Ji, Jj;
  double h = 1E-12;

  residualFunction(residual0, stepsize);

  for(list<Element *>::iterator it = dynamicElements.begin();
      it != dynamicElements.end(); it++)
  {
    Element &e = **it;

    ej = e.index;

    set<Element *> dependents;

    elementDependents(e, dependents);

    for(int comp = 0; comp < 6; comp++)
    {
      if(comp < 3)
      {
	e.position(comp) += h;
	j = 3 * ej + comp;
      }
      else
      {
	e.velocity(comp - 3) += h;
	j = 3 * n + 3 * ej + comp - 3;
      }
      
      for(set<Element *>::iterator it2 = dependents.begin();
	  it2 != dependents.end(); it2++)
      {
	Element &e2 = **it2;
	
	ei = e2.index;

	residualFunction(residual1, stepsize, &e2);

	Ji = 3 * ei;

	J(Ji + 0, j) = (residual1(Ji + 0) - residual0(Ji + 0)) / h;
	J(Ji + 1, j) = (residual1(Ji + 1) - residual0(Ji + 1)) / h;
	J(Ji + 2, j) = (residual1(Ji + 2) - residual0(Ji + 2)) / h;

	Ji = 3 * n + 3 * ei;

	J(Ji + 0, j) = (residual1(Ji + 0) - residual0(Ji + 0)) / h;
	J(Ji + 1, j) = (residual1(Ji + 1) - residual0(Ji + 1)) / h;
	J(Ji + 2, j) = (residual1(Ji + 2) - residual0(Ji + 2)) / h;

      }

      if(comp < 3)
      {
	e.position(comp) -= h;
      }
      else
      {
	e.velocity(comp - 3) -= h;
      }
    }
    
    
  }

  cerr << "nderivs: " << nderivs << endl;
}

void BasicEulerIntegrator::elementDependents(Element &e, set<Element *> &dependents)
{
  // e depends on itself

  dependents.insert(&e);

  // e depends on its connectants
    
  for(list<Connection *>::iterator it = e.connections.begin();
      it != e.connections.end(); it++)
  {
    Element *e2;
    
    if(&e == (*it)->element1)
      e2 = (*it)->element2;
    else
      e2 = (*it)->element1;
    
    if(!e2->artificial)
    {
      dependents.insert(e2);
    }
  }
}

}

