#include <stdio.h>

#include <strstream>
#include <iostream>
#include <iomanip>
#include <map>

#include <sys/time.h>

#include "ko/Simulator.hpp"

using namespace std;

namespace ko
{

Simulator::Simulator(Physical &physical, double duration) :
  physical(physical), time(0), stepsize(0.01), duration(duration),
  euler(physical), multiAdaptive(physical, duration),
  doCollisions(true),
  //integrator(multiAdaptive)
  integrator(&euler)
{

}

void Simulator::step()
{
  double origstep = stepsize;
  bool adequateSolution = false;

  integrator->fixedk = stepsize;

  while(!adequateSolution)
  {

    double elapsed, t1;
    struct timeval before, after;
    //gettimeofday(&before, NULL);


    adequateSolution = integrator->integrate(time, t1);

    if(!adequateSolution)
    {
      cerr << "fixedk = " << integrator->fixedk <<
	" did not converge, halving k" << endl;

      //integrator->fixedk /= 2;
      integrator->fixedk *= 0.5;
    }
    else
    {

      time = t1;

      //gettimeofday(&after, NULL);
      
      //elapsed = (1E6 * after.tv_sec + after.tv_usec) -
      //(1E6 * before.tv_sec + before.tv_usec);
      
      //cerr << "integrate: " << elapsed << endl;
      
      
      // Don't do collisions for now
      
      //adequateSolution = true;
      
      if(doCollisions)
	collisions();
      newConnections.clear();
    }
  }

  cerr << "time: " << time << endl;

  if(doCollisions)
    fractures();
  
  //plasticity();

  //time += stepsize;
  //stepsize = origstep;

  stepsize = integrator->fixedk;
}




bool Simulator::collisions()
{
  double elapsed;
  struct timeval before, after;

  gettimeofday(&before, NULL);

  //buildTrees();
  idetector.init(physical.elements);
  
  for(list<Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    Element &e1 = **it;
      
    list<Element *> collisions;

    idetector.intersectionList(e1, collisions);

    for(list<Element *>::iterator it2 = collisions.begin();
	it2 != collisions.end(); it2++)
    {
      Element &e2 = **it2;

      if(isCollision(e1, e2))
      {
	cerr << "collision!" << endl;
	
	Connection &c = *(new Connection());
	
	c.connect(e1, e2);
	c.transient = true;
	
	Vector r(3);
	
	r = e2.position;
	r.axpy(-1, e1.position);
	
	c.length = e1.radius + e2.radius;
	
	physical.addConnection(c);

	newConnections.push_back(&c);

	cout << "contact connection: " << endl;

	cout << "e1: " << e1.name << endl;
	cout << "e2: " << e2.name << endl;
	
	cout << "hooke: " << MechanicsModel::hooke(c) << endl;
	cout << "damp: " << MechanicsModel::damp(c) << endl;


	// Check error

	Vector vrel(3);
    
	vrel = e1.velocity;
	vrel.axpy(-1, e2.velocity);
	
	if(vrel.norm() * stepsize > min(e1.radius, e2.radius) / 1)
	{
	  cerr << "large error collision predicted: " <<
	    vrel.norm() * stepsize / min(e1.radius, e2.radius) << endl;
	  return false;

	}
      }
    }
  }

  gettimeofday(&after, NULL);
  
  elapsed = (1E6 * after.tv_sec + after.tv_usec) -
    (1E6 * before.tv_sec + before.tv_usec);
  
  cerr << "collisions: " << elapsed << endl;
  return true;
}

void Simulator::fractures()
{
  list<Connection *> removals;

  for(list<Connection *>::iterator it = physical.connections.begin();
      it != physical.connections.end(); it++)
  {
    Connection &c = **it;

    double strength;

    if(c.transient)
    {
//       cerr << "transient" << endl;
      strength = (c.element1->material->collisionStrength +
		  c.element2->material->collisionStrength) / 2 / c.damage;
    }
    else
    {
      strength = (c.element1->material->strength +
		  c.element2->material->strength) / 2 /	c.damage;
    }

    //cerr << "tension: " << c.tension << endl;
    //cerr << c.element1->name << " <-> " << c.element2->name << endl;

    if(c.tension > strength)
    {
//       cerr << "fracture: " << c.tension << " " << strength << " " << c.damage << endl;
      //c.disconnect();
      removals.push_back(&c);
    }


    if(c.transient)
    {
//       cerr << "disconnect: " << strength << " " << c.tension << endl;
    }
  }

  for(list<Connection *>::iterator it = removals.begin();
      it != removals.end(); it++)
  {
    Connection &c = **it;

    physical.removeConnection(c);
  }
}

void Simulator::plasticity()
{
  /*
  for(list<Connection *>::iterator it = physical.connections.begin();
      it != physical.connections.end(); it++)
  {
    Connection &c = **it;

    mmodel.plasticity(c);
  }  
  */
}

void Simulator::insertPhysical(Physical &p)
{
  physical.merge(p);
}


bool Simulator::isCollision(Element &e1, Element &e2)
{
  double rnorm;
  
  Vector r(3);
  
  r = e1.position;
  r.axpy(-1, e2.position);
  
  rnorm = r.norm();
  
  if(rnorm <= (e1.radius + e2.radius))
  {
    Vector vrel(3);
    
    double vdir;
    
    vrel = e1.velocity;
    vrel.axpy(-1, e2.velocity);
    
    // project vrel onto connection
    vrel.axpy((vrel * r) / (rnorm * rnorm), r);
    
    // how do the velocities point?
    vdir = (r * vrel);
    
    //if(e1.name == "element 1000" || e2.name == "element 1000")
//     cerr << "vdir: " << vdir << endl;
    
    // If the velocities are pointing away from each other, we say there
    // is no collision.
    if(true || vdir < 0)
    {
      return true;
    }
  }
  return false;
}

}

