// Copyright (C) 2004 Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Modified by Johan Jansson, 2004.

// Benchmark problem for the multi-adaptive ODE-solver, a system of n
// particles connected with springs. All springs, except the first spring,
// are of equal stiffness k = 1.

#include <iostream>
#include <dolfin.h>

using namespace dolfin;

class Benchmark : public ParticleSystem
{
public:
  
  Benchmark(unsigned int n) : ParticleSystem(n, 3)
  {
    // Final time
    T = 100.0;

    // Spring constant
    k = 1.0;
    
    // Grid size
    h = 1.0 / static_cast<real>(n - 1);
    //h = 1.0;

    // Compute sparsity
    sparse();
  }

  real x0(unsigned int i)
  {
    if ( i == 0 )
      return h/2;

    return h * static_cast<real>(i);
  }

  real y0(unsigned int i)
  {
    return 0.0;
  }

  real z0(unsigned int i)
  {
    return 0.0;
  }

  real vx0(unsigned int i)
  {
    return 0.0;
  }

  real vy0(unsigned int i)
  {
    return 0.0;
  }

  real vz0(unsigned int i)
  {
    return 0.0;
  }

  real Fx(unsigned int i, real t)
  {
    // Adapted from Ko

    for(int ii = 0; ii < 2; ii++)
    {

      // Do the operations performed in Ko to put the work done in "f"
      // on par.

      static Vector r(3), vrel(3), force(3), vrelNormal(3), vrelParallel(3),
	e1position(3), e1velocity(3), e2position(3), e2velocity(3);
      
      real hooke = 1.0, length = 1.0, tension, rnorm, vrelnorm;
      
      e1position(0) = x(i);
      e1position(1) = y(i);
      e1position(2) = z(i);
      
      e2position(0) = x(i + 1);
      e2position(1) = y(i + 1);
      e2position(2) = z(i + 1);
      
      e1velocity(0) = x(i);
      e1velocity(1) = y(i);
      e1velocity(2) = z(i);
      
      e2velocity(0) = x(i + 1);
      e2velocity(1) = y(i + 1);
      e2velocity(2) = z(i + 1);
      
      force = 0;
      
      r = e1position;
      r.axpy(-1, e2position);
      
      vrel = e1velocity;
      vrel.axpy(-1, e2velocity);
      
      rnorm = r.norm();
      vrelnorm = vrel.norm();
      
      vrelParallel = r;
      vrelParallel *= (vrel * r) / (rnorm * rnorm);
      
      force = r;
      force *= -hooke * (rnorm - length) / rnorm;
      
      if((r * force) > 0)
	tension = -2 * force.norm() / length;
      else
	tension = 2 * force.norm() / length;
    }
    
    if ( i == 0 )
      return - 100.0*k*x(i) + k*(x(i+1) - x(i) - h);
    else if ( i == (n-1) )
      return - k*(x(i) - x(i-1) - h) + k*(1.0 - x(i));
    else
      return - k*(x(i) - x(i-1) - h) + k*(x(i+1) - x(i) - h);
  }
  
  real Fy(unsigned int i, real t)
  {
    // Adapted from Ko

    for(int ii = 0; ii < 2; ii++)
    {

      // Do the operations performed in Ko to put the work done in "f"
      // on par.

      static Vector r(3), vrel(3), force(3), vrelNormal(3), vrelParallel(3),
	e1position(3), e1velocity(3), e2position(3), e2velocity(3);
      
      real hooke = 1.0, length = 1.0, tension, rnorm, vrelnorm;
      
      e1position(0) = x(i);
      e1position(1) = y(i);
      e1position(2) = z(i);
      
      e2position(0) = x(i + 1);
      e2position(1) = y(i + 1);
      e2position(2) = z(i + 1);
      
      e1velocity(0) = x(i);
      e1velocity(1) = y(i);
      e1velocity(2) = z(i);
      
      e2velocity(0) = x(i + 1);
      e2velocity(1) = y(i + 1);
      e2velocity(2) = z(i + 1);
      
      force = 0;
      
      r = e1position;
      r.axpy(-1, e2position);
      
      vrel = e1velocity;
      vrel.axpy(-1, e2velocity);
      
      rnorm = r.norm();
      vrelnorm = vrel.norm();
      
      vrelParallel = r;
      vrelParallel *= (vrel * r) / (rnorm * rnorm);
      
      force = r;
      force *= -hooke * (rnorm - length) / rnorm;
      
      if((r * force) > 0)
	tension = -2 * force.norm() / length;
      else
	tension = 2 * force.norm() / length;
    }
    

    return 0.0;
  }

  real Fz(unsigned int i, real t)
  {
    // Adapted from Ko

    for(int ii = 0; ii < 2; ii++)
    {

      // Do the operations performed in Ko to put the work done in "f"
      // on par.

      static Vector r(3), vrel(3), force(3), vrelNormal(3), vrelParallel(3),
	e1position(3), e1velocity(3), e2position(3), e2velocity(3);
      
      real hooke = 1.0, length = 1.0, tension, rnorm, vrelnorm;
      
      e1position(0) = x(i);
      e1position(1) = y(i);
      e1position(2) = z(i);
      
      e2position(0) = x(i + 1);
      e2position(1) = y(i + 1);
      e2position(2) = z(i + 1);
      
      e1velocity(0) = x(i);
      e1velocity(1) = y(i);
      e1velocity(2) = z(i);
      
      e2velocity(0) = x(i + 1);
      e2velocity(1) = y(i + 1);
      e2velocity(2) = z(i + 1);
      
      force = 0;
      
      r = e1position;
      r.axpy(-1, e2position);
      
      vrel = e1velocity;
      vrel.axpy(-1, e2velocity);
      
      rnorm = r.norm();
      vrelnorm = vrel.norm();
      
      vrelParallel = r;
      vrelParallel *= (vrel * r) / (rnorm * rnorm);
      
      force = r;
      force *= -hooke * (rnorm - length) / rnorm;
      
      if((r * force) > 0)
	tension = -2 * force.norm() / length;
      else
	tension = 2 * force.norm() / length;
    }
    

    return 0.0;
  }

private:

  real k;
  real h;
  
};

int main()
{
  //double k = 0.03334 / 10.0;

  dolfin_set("output", "plain text");
  dolfin_set("tolerance", 0.01);
  dolfin_set("number of samples", 100);
  dolfin_set("solve dual problem", false);

  //dolfin::dolfin_set("initial time step", k);
  //dolfin::dolfin_set("fixed time step", true);

  Benchmark bench(100);
  //Benchmark bench(1);
  bench.solve();
  
  return 0;
}
