// Copyright (C) 2004 Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Modified by Johan Jansson, 2004.

// Benchmark problem for the multi-adaptive ODE-solver, a system of n
// particles connected with springs. All springs, except the first spring,
// are of equal stiffness k = 1.

#include <iostream>
#include <dolfin.h>

using namespace dolfin;


class Benchmark : public ParticleSystem
{
public:
  
  Benchmark(unsigned int n) : ParticleSystem(n, 3)
  {
    if ( n < 2 )
      dolfin_error("System must have at least 2 particles.");

    // Final time
    T = 100.0;

    // Spring constant
    k = 10.0;

    // Damp constant
    b = 0 * 1.0; 
    
    // Grid size
    h = 0.1;
    //h = 1.0;

    // Compute sparsity
    sparse();
  }

  void computeForce(Vector &e0position, Vector &e1position, Vector &e0velocity,
		    Vector &e1velocity, Vector &force)
  {
    static Vector r(3), vrel(3), vrelParallel(3);
    real rnorm;
    
    // Force between the mass elements
    r = e0position;
    r.add(-1, e1position);
    
    rnorm = r.norm();

    vrel = e0velocity;
    vrel.add(-1, e1velocity);

    vrelParallel = r;
    vrelParallel *= (vrel * r) / (rnorm * rnorm);

    //std::cerr << "vrelParallel: " <<
    //  vrelParallel(0) << " " <<
    //  vrelParallel(1) << " " <<
    //  vrelParallel(2) << std::endl;
    
    // Hooke
    force = r;
    force *= -k * (rnorm - 0.1) / rnorm;

    force.add(-b, vrelParallel);
  }

  real x0(unsigned int i)
  {
    if ( i == 0 )
      return 0.1;
    if ( i == 1 )
      return 0.1;

    return 0;
  }

  real y0(unsigned int i)
  {
    if ( i == 0 )
      return -1e-2;
    //return 0;
    if ( i == 1 )
      return 1e-1;

    return 0.0;
  }

  real z0(unsigned int i)
  {
    return 0.0;
  }

  real vx0(unsigned int i)
  {
    return 0.0;
  }

  real vy0(unsigned int i)
  {
    return 0.0;
  }

  real vz0(unsigned int i)
  {
    return 0.0;
  }

  real Fx(unsigned int i, real t)
  {
    // Adapted from Ko

    static Vector e0position(3), e1position(3), r(3), force(3),
      fixed0(3), fixed1(3), fixedforce0(3), fixedforce1(3);

    static Vector e0velocity(3), e1velocity(3);

    static Vector zero(3);
    zero = 0;

    e0position(0) = x(0);
    e0position(1) = y(0);
    e0position(2) = z(0);

    e1position(0) = x(1);
    e1position(1) = y(1);
    e1position(2) = z(1);

    e0velocity(0) = vx(0);
    e0velocity(1) = vy(0);
    e0velocity(2) = vz(0);

    e1velocity(0) = vx(1);
    e1velocity(1) = vy(1);
    e1velocity(2) = vz(1);

    fixed0(0) = 0;
    fixed0(1) = 0;
    fixed0(2) = 0;

    fixed1(0) = 0;
    fixed1(1) = 0.1;
    fixed1(2) = 0;

    computeForce(e0position, e1position, e0velocity, e1velocity, force);
    computeForce(e0position, fixed0, e0velocity, zero, fixedforce0);
    computeForce(e1position, fixed1, e1velocity, zero, fixedforce1);

    if(i == 0)
    {
      return force(0) + fixedforce0(0);
      //return force(0);
    }
    if(i == 1)
    {
      return -force(0) + fixedforce1(0);
      //return -force(0);
    }
    return 0.0;

  }
  
  real Fy(unsigned int i, real t)
  {
    // Adapted from Ko

    static Vector e0position(3), e1position(3), r(3), force(3),
      fixed0(3), fixed1(3), fixedforce0(3), fixedforce1(3);

    static Vector e0velocity(3), e1velocity(3);

    static Vector zero(3);
    zero = 0;

    e0position(0) = x(0);
    e0position(1) = y(0);
    e0position(2) = z(0);

    e1position(0) = x(1);
    e1position(1) = y(1);
    e1position(2) = z(1);

    e0velocity(0) = vx(0);
    e0velocity(1) = vy(0);
    e0velocity(2) = vz(0);

    e1velocity(0) = vx(1);
    e1velocity(1) = vy(1);
    e1velocity(2) = vz(1);

    fixed0(0) = 0;
    fixed0(1) = 0;
    fixed0(2) = 0;

    fixed1(0) = 0;
    fixed1(1) = 0.1;
    fixed1(2) = 0;

    computeForce(e0position, e1position, e0velocity, e1velocity, force);
    computeForce(e0position, fixed0, e0velocity, zero, fixedforce0);
    computeForce(e1position, fixed1, e1velocity, zero, fixedforce1);

    if(i == 0)
    {
      return force(1) + fixedforce0(1);
      //return force(1);
    }
    if(i == 1)
    {
      return -force(1) + fixedforce1(1);
      //return -force(1);
    }
    return 0.0;
  }

  real Fz(unsigned int i, real t)
  {
    // Adapted from Ko

    static Vector e0position(3), e1position(3), r(3), force(3),
      fixed0(3), fixed1(3), fixedforce0(3), fixedforce1(3);

    static Vector e0velocity(3), e1velocity(3);

    static Vector zero(3);
    zero = 0;

    e0position(0) = x(0);
    e0position(1) = y(0);
    e0position(2) = z(0);

    e1position(0) = x(1);
    e1position(1) = y(1);
    e1position(2) = z(1);

    e0velocity(0) = vx(0);
    e0velocity(1) = vy(0);
    e0velocity(2) = vz(0);

    e1velocity(0) = vx(1);
    e1velocity(1) = vy(1);
    e1velocity(2) = vz(1);

    fixed0(0) = 0;
    fixed0(1) = 0;
    fixed0(2) = 0;

    fixed1(0) = 0;
    fixed1(1) = 0.1;
    fixed1(2) = 0;

    computeForce(e0position, e1position, e0velocity, e1velocity, force);
    computeForce(e0position, fixed0, e0velocity, zero, fixedforce0);
    computeForce(e1position, fixed1, e1velocity, zero, fixedforce1);

    if(i == 0)
    {
      return force(2) + fixedforce0(2);
      //return force(2);
    }
    if(i == 1)
    {
      return -force(2) + fixedforce1(2);
      //return -force(2);
    }
    return 0.0;
  }

private:

  real k;
  real h;
  real b;
  
};

int main()
{
  //double k = 0.03334 / 10.0;
  double k = 0.01;

  dolfin_set("output", "plain text");
  //dolfin_set("tolerance", 0.0001);
  dolfin_set("tolerance", 0.001);
  dolfin_set("number of samples", 1000);
  dolfin_set("solve dual problem", false);
  dolfin_set("initial time step", k);
  dolfin_set("fixed time step", true);
  dolfin::dolfin_set("method", "dg");                                       
  dolfin::dolfin_set("order", 0);                                           


  Benchmark bench(2);
  //Benchmark bench(1);
  bench.solve();
  
  return 0;
}
