// Copyright (C) 2004-2005 Johan Jansson.
// Licensed under the GNU GPL Version 2.
//
// First added:  2004
// Last changed: 2005

#include <strstream>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>

#include <dolfin.h>
#include <dolfin/constants.h>
#include <ko/Element.hpp>
#include <ko/Connection.hpp>
#include <ko/Physical.hpp>
#include <ko/Simulator.hpp>
#include <ko/PhysicalDolfinInterface.hpp>
#include <ko/PhysicalDOMTranslator.hpp>
#include <ko/IntersectionDetector.hpp>


using namespace dolfin;

class Source;
class InitialVelocity;
class Density;

class KoSolver : public ElasticityUpdatedSolver
{
public:
  KoSolver(Mesh& mesh,
	   Source& f, InitialVelocity& v0, Density& rho,
	   real& E, real& nu, real& nuv, real& nuplast,
	   BoundaryCondition& bc, real& k, real& T);
  virtual void save(Mesh& mesh, File &solutionfile, real t);
  void preparestep();
  void prepareiteration();
  void buildIntersections();

  // Data

  ko::MechanicsModel mmodel;
  ko::Physical physical;
  ko::PhysicalDolfinInterface interface;
  ko::PhysicalDOMTranslator domtranslator;
  ko::IntersectionDetector idetector;
  ko::Simulator simulator;

  int contactcounter;
  real lastcontactsample;

//   std::vector< std::vector<Node *> > intersections;
  std::vector< std::vector<Node *> > intersections;

};

// Density
class Density : public Function
{
  real operator() (const Point& p) const
  {
    if(p.x <= 0.0 && p.y >= 0.0 && p.y <= 2.0)
    {
      return 1.0e3;
    }
    else if(p.y < -10.0)
    {
      return 1.0e30;
    }
    else
      return 1.0e3;
  }
};

// Right-hand side
class Source : public Function
{
public:
  Source() : kosolver(0)
  {
  }
  real operator() (const Point& p, unsigned int i) const
  {
//      cout << "cell id: " << _cell->id() << endl;
//     cout << "mesh: " << _mesh << endl;

    real result = 0;

//     // Contact forces
//     // The idea is that not many elements need to compute contact forces,
//     // so the computation does not need aggressive optimization
    
//     int cid = _cell->id();
//     std::vector<Node *>& contacts = kosolver->intersections[cid];
    
//     if(contacts.size() != 0)
//     {

// //       cout << "computing contact: " << cid << endl;

//       dolfin::real h = _cell->diameter();
// //       dolfin::real h = 3.0;
      
// //       dolfin::real E = 1.0e3 * kosolver->E;
//       dolfin::real E = 200 * kosolver->E;
// //       dolfin::real E = 20 * kosolver->E;
//       dolfin::real nuv = 1 * kosolver->nuv;
//       dolfin::real friction = 0.0;
      
//       // Find out which node p corresponds to (assumes Lagrange P1)
//       // FIXME: Need function interpolation for general case

//       dolfin::Node* np = 0;

//       for(dolfin::NodeIterator n(*_cell); !n.end(); ++n)
//       {
// 	dolfin::Node& node = *n;
	
// 	if(p.dist(node.coord()) == 0.0)
// 	{
// 	  np = &node;
// 	}
//       }

//       // Use node id to compute dof

//       if(np == 0)
//       {
// 	cout << "Source: couldn't find node, exiting" << endl;
// 	exit(1);
//       }
	
//       dolfin::Node& node1 = *np;
//       int nid = node1.id();

//       // Evaluate quantities for p

//       ko::SimpleVector force, pforce, v1, p1;

//       Function& v = kosolver->v1;

//       p1(0) = node1.coord().x;
//       p1(1) = node1.coord().y;
//       p1(2) = node1.coord().z;
      
//       v1(0) = v(node1, 0);
//       v1(1) = v(node1, 1);
//       v1(2) = v(node1, 2);

//       ko::SimpleVector cforce;

//       for(std::vector<Node *>::iterator it = contacts.begin();
// 	  it != contacts.end(); it++)
//       {
// 	dolfin::Node& node2 = **it;
// 	dolfin::Point& point2 = node2.coord();
// 	dolfin::real dist = p.dist(point2);
	
// 	dolfin::real hcont = node2.cell(0).diameter();
// 	dolfin::real l = 0.40 * (h + hcont);
	
// // 	cout << "checking contact:" << endl;
// // 	cout << "cell id: " << cid << endl;	
// // 	cout << "h: " << h << endl;	
// // 	cout << "hcont: " << hcont << endl;	
// // 	cout << p << endl;
// // 	cout << "dist: " << dist << endl;	
// // 	cout << "l: " << l << endl;	

// 	if(dist < l)
// 	{
// // 	  cout << "contact:" << endl;
// // 	  cout << "cell id: " << cid << endl;
// // 	  cout << p << endl;
// // 	  cout << "node id: " << node2.id() << endl;
// // 	  cout << node2.coord() << endl;
	  
//  	  ko::SimpleVector force, r, p2, v2, vrel, vrelParallel, vrelNormal,
// 	    tmp, normf;
// 	  dolfin::real rnorm;
	  
// 	  p2(0) = node2.coord().x;
// 	  p2(1) = node2.coord().y;
// 	  p2(2) = node2.coord().z;

// 	  v2(0) = v(node2, 0);
// 	  v2(1) = v(node2, 1);
// 	  v2(2) = v(node2, 2);

// 	  r = p1;
// 	  r.axpy(-1, p2);
// 	  rnorm = r.norm();

// // 	  cout << "rnorm: " << rnorm << endl;
// // 	  cout << "r(0): " << r(0) << endl;
// // 	  cout << "r(1): " << r(1) << endl;
// // 	  cout << "r(2): " << r(2) << endl;

// 	  vrel = v1;
// 	  vrel.axpy(-1, v2);
	  
// 	  vrelParallel = r;
// 	  vrelParallel *= (vrel * r) / (rnorm * rnorm);

// 	  vrelNormal = vrel;
// 	  vrelNormal.axpy(-1, vrelParallel);

// 	  // Hooke

// 	  tmp = r;
// 	  // 	    cout << "tmp(0): " << tmp(0) << endl;
// 	  // 	    cout << "tmp(1): " << tmp(1) << endl;
// 	  // 	    cout << "tmp(2): " << tmp(2) << endl;
// 	  tmp *= E * (l - rnorm) / rnorm;
// 	  force += tmp;
	  
// // 	  // Damping

// // 	  tmp = 0;
// // 	  tmp.axpy(-nuv, vrelParallel);

// // 	  force += tmp;

// // 	  // Friction
// // 	  tmp = vrelNormal;
// // 	  tmp *= -friction * force.norm();
// // 	  force += tmp;

//  	  cforce = force;
// 	}
//       }

//       result += cforce(i);
//     }


    // Gravitational forces
    
//     dolfin::real density = kosolver->rho(p);

    if(i == 1)
    {
      result += 1.0 * -9.81 * 1.0e3;
    }
    
    return result;
  }

  KoSolver* kosolver;
};



// Initial velocity
class InitialVelocity : public Function
{
  real operator() (const Point& p, unsigned int i) const
  {
    real result = 0.0;
    
//     if(i == 1 && p.y > -100.0 && p.y < 1.0 && p.z > 0.0 && (p.x > 0.0 || p.x < 0.0))
//     {
//       result += -2.0;
//     }

//     if(i == 1 && p.x < 2.0)
//     {
//       result += 3.0;
//     }
//     if(i == 0 && p.y >= 0.0)
//     {
//       result += 9.0;
//     }

//     if(i == 0 && p.y >= -2.0)
//     {
//       result += 5.0;
//     }

//     if(i == 1 && p.x >= 0.0)
//     {
//       result += 1.0;
//     }

//     if(i == 0)
//     {
//       if(p.y > 1.0)
// 	result += -1.5;
//       if(p.y < 1.0)
// 	result += 1.5;
//     }
//     if(i == 1)
//     {
//       if(p.x > 1.0)
// 	result += 1.5;
//       if(p.x < 1.0)
// 	result += -1.5;
//     }


//     if(i == 1 && p.y > 1.0)
//     {
//       return 10.0;
//     }
//     if(i == 0 && p.y > 1.0)
//     {
//       return -0.6;
//     }
//     if(i == 1 && p.y <= 1.0 && p.y >= 0.0)
//     {
//       return -0.5;
//     }


//     Point w, r, v, center;

//     center.x = 1.0;
//     center.y = 1.0;
//     center.z = 0.0;

//     r = p - center;

//     w.x = 0.0;
//     w.y = 0.0;
//     w.z = 2.0;

//     v = w.cross(r);

//     if(i == 0)
//     {
//       result += v.x;
//     }
//     if(i == 1)
//     {
//       result += v.y;
//     }

//     if(i == 1 && p.x <= 0.5 && p.y >= 0.0)
//     {
//       result += 0.5;
//     }

//     if(i == 0 && p.y >= -10.0)
//     {
//       result += 10.0;
//     }


    return result;
  }
};

// Boundary condition
class MyBC : public BoundaryCondition
{
public:
  MyBC::MyBC() : BoundaryCondition()
  {
  }

  const BoundaryValue operator() (const Point& p, unsigned int i)
  {
    BoundaryValue value;

//     if(p.y < -100.0 || p.y > 1.0 || p.z < 0.0 || p.x == 0.0)
//     if(p.y < -100.0 || p.x <= -0.5)
    if(p.y < -100.0)
    {
      value.set(0.0);
    }    


    return value;
  }
};

KoSolver::KoSolver(Mesh& mesh,
		   Source& f, InitialVelocity& v0, Density& rho,
		   real& E, real& nu, real& nuv, real& nuplast,
		   BoundaryCondition& bc, real& k, real& T) :
  ElasticityUpdatedSolver(mesh, f, v0, rho, E, nu, nuv, nuplast, bc, k, T),
  interface(mesh, physical, v1, E, nuv), simulator(physical, T),
  contactcounter(0), lastcontactsample(0)
{
  idetector.init(physical.elements);
  
  intersections.resize(mesh.noCells());

  f.kosolver = this;
}

void KoSolver::save(Mesh& mesh, File &solutionfile, real t)
{
  ElasticityUpdatedSolver::save(mesh, solutionfile, t);
  
  std::cerr << "Ko saving" << std::endl;
  
  std::ostrstream filename;
  
  filename.fill('0');
  filename.width(6);
  
  filename << counter << ".xml" << std::ends;
  
  char *fname = filename.str();
  
  std::cerr << "writing" << std::endl;
  std::cerr << fname << std::endl;
  
  std::ofstream os;
  os.open(fname, std::ios::out);
  
  domtranslator.write(os, physical);
}

void KoSolver::preparestep()
{
  real contactsamplefreq = 1 * k;
  
  while(lastcontactsample + contactsamplefreq < t || t == 0.0)
  {
    std::cerr << "prepare step" << std::endl;
    
//     idetector.buildTrees();
    
//     buildIntersections();
    simulator.collisions();
    simulator.fractures();
    
    contactcounter++;
    lastcontactsample = std::min(t, lastcontactsample + contactsamplefreq);
    
    if(t == 0.0)
    {
      break;
    }
  }
}

void KoSolver::prepareiteration()
{
  interface.updatePhysical();

  fcontact = 0.0;

  for(std::list<ko::Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    ko::Element *e = *it;
    int i = e->index;

    e->force = 0.0;
  }

  for(std::list<ko::Connection *>::iterator it = physical.connections.begin();
      it != physical.connections.end(); it++)
  {
    ko::Connection &c = **it;
    if(c.transient)
    {
      static ko::Vector force(3);
      
      ko::Element *e1 = c.element1;
      ko::Element *e2 = c.element2;
      
      mmodel.force(c, *e1, force);
      e1->force += force;
      force *= -1;
      e2->force += force;
      
      // Project force on connection
      static ko::Vector r(3), tmp(3);
      double rnorm;
      
      r = e1->position;
      r.axpy(-1, e2->position);
      
      rnorm = r.norm();
      
      tmp = r;
      tmp *= (force * r) / (rnorm * rnorm);
      
      c.tension = tmp.norm();
      
      if(force * r < 0)
	c.tension *= -1;
      
//       cout << "tension: " << c.tension << endl;
    }
  }

  int offset = mesh.noNodes();
  for(std::list<ko::Element *>::iterator it = physical.elements.begin();
      it != physical.elements.end(); it++)
  {
    ko::Element *e = *it;
    int i = e->index;

    fcontact(0 * offset + i) = e->force(0); 
    fcontact(1 * offset + i) = e->force(1); 
    fcontact(2 * offset + i) = e->force(2); 
  }
}

void KoSolver::buildIntersections()
{


//   for(dolfin::CellIterator c(mesh); !c.end(); ++c)
//   {
//     dolfin::Cell &cell = *c;
//     int cid = cell.id();
    
//     intersections[cid].clear();
//   }

//   for(dolfin::NodeIterator nmesh(mesh); !nmesh.end(); ++nmesh)
//   {
//     dolfin::Node& node1 = *nmesh;
//     int id1 = node1.id();
//     ko::Element& e1 = *(interface.elements[id1]);
    
//     std::list<ko::Element *> ilist;
//     idetector.intersectionList(e1, ilist);
    
//     for(std::list<ko::Element *>::iterator it = ilist.begin();
// 	it != ilist.end(); it++)
//     {
//       ko::Element& e2 = **it;
//       Node& node2 = *(interface.nodes[e2.index]);
//       int id2 = node2.id();

//       cout << "intersection: " << id1 << " " << id2 << endl;

//       for(dolfin::CellIterator c(node1); !c.end(); ++c)
//       {
// 	dolfin::Cell &cell = *c;

// 	intersections[cell.id()].push_back(&node2);
//       }
//     }
//   }


}


int main(int argC, char **argV)
{
  dolfin_output("plain text");
//   dolfin_output("silent");
  dolfin_set("progress step", 1.0);

  std::string physicalfile, dolfinfile;
  
  dolfin::dolfin_output("plain text");
  
  dolfin::real T = 10.0;
  dolfin::real g = -9.81;
  std::string method = "stab";
  
  if(argC > 1)
  {
    physicalfile = argV[1];
  }
  
  if(argC > 2)
  {
    dolfinfile = argV[2];
    
    dolfin::dolfin_load(dolfinfile.c_str());
    std::cerr << "foo" << std::endl;
    
    std::cerr << "tolerance: " << (double)dolfin::dolfin_get("tolerance") << std::endl;
    //cerr << "method: " << (char *)dolfin::dolfin_get("method") << std::endl;
  }
  
  if(argC > 3)
  {
    std::string arg = argV[3];
    if(arg == "stab")
      method = arg;
    if(arg == "newton")
      method = arg;
    if(arg == "nostab")
      method = arg;
    
  }
  
  std::cerr << "method: " << method << std::endl;

  if(argC > 4)
  {
    T = atof(argV[4]);

  }

  std::cerr << "T: " << T << std::endl;

  if(argC > 5)
  {
    g = atof(argV[5]);

  }

  std::cerr << "g: " << g << std::endl;

//   dolfin::real k = dolfin::dolfin_get("maximum time step");;
  dolfin::real k = 1e-4;

  dolfin_set("method", "cg");
  dolfin_set("order", 1);
  dolfin_set("save solution", false);
  dolfin_set("fixed time step", true);
  dolfin_set("initial time step", k);
  dolfin_set("maximum time step", k);
  dolfin_set("discrete tolerance", 1.0e-4);



  //Mesh mesh("cow01.xml.gz");
  //Mesh mesh("cow07.xml.gz");
  //Mesh mesh("mymesh01.xml.gz");
//   Mesh mesh("collision05.xml.gz");
//   UnitCube mesh(4, 4, 4);
  //Mesh mesh("roterror01.xml.gz");

//    Mesh mesh("collision16.xml.gz");
//    Mesh mesh2("collision17.xml.gz");

   Mesh mesh("floor01.xml.gz");
//    Mesh mesh2("cow05.xml.gz");
//   Mesh mesh3("cow05b.xml.gz");
//   Mesh mesh2("block01.xml.gz");
   //Mesh mesh2("block03a.xml.gz");
   //Mesh mesh3("block03b.xml.gz");
   Mesh mesh2("minimal4.xml.gz");

//   mesh2.refineUniformly();
//   mesh3.refineUniformly();

   mesh.merge(mesh2);
//   mesh.merge(mesh3);

//   for(dolfin::NodeIterator nmesh(mesh); !nmesh.end(); ++nmesh)
//   {
//     dolfin::Node& node1 = *nmesh;


//     node1.coord().y += 3.0;
//   }

//   File meshfile("mymesh.xml.gz");

//   meshfile << mesh;

  Boundary boundary(mesh);
  //  Mesh mesh("tetmesh-1c.xml.gz");

//   for (NodeIterator n(&mesh); !n.end();)
//   {
//     Node &cnode = *n;
//     ++n;

//     if(cnode.coord().y > 10.0)
//     {
//       mesh.remove(cnode);
//     }
//   }

//   mesh.init();

//   File outfile("unitcube-07.xml.gz");
//   outfile << mesh;
  
  Source f;
  Density rho;
  InitialVelocity v0;
  MyBC bc;

  real E = 2 * 1.0e5; // Young's modulus
  real nu = 0.3; // Poisson's ratio
  real nuv = 4 * 1e2; // viscosity
  real nuplast = 0.0; // viscosity

//   real E = 1 * 1.0e5; // Young's modulus
//   real nu = 0.3; // Poisson's ratio
//   real nuv = 1.0 * 1e3; // viscosity
//   real nuplast = 0.0; // viscosity

//   real E = 1 * 4.0e5; // Young's modulus
//   real nu = 0.3; // Poisson's ratio
//   real nuv = 1.0 * 1e3; // viscosity
//   real nuplast = 0.0; // viscosity

//   real E = 5 * 5.0e4; // Young's modulus
//   real nu = 0.3; // Poisson's ratio
//   real nuv = 5.0 * 1e3; // viscosity
//   real nuplast = 0.0; // viscosity

//   ElasticityUpdatedSolver::solve(mesh, f, v0, rho, E, nu, nuv, bc, k, T);
  KoSolver kosolver(mesh, f, v0, rho, E, nu, nuv, nuplast, bc, k, T);
//   kosolver.rtol = 1e-10;
//   kosolver.savesamplefreq = 1.0 / k / 10.0;

  // Drastically improve contact stability by using a uniform mass matrix
  // Should be density * h / 2
  kosolver.m = 2.0;

//   kosolver.interface.m.young = 0.2 * E;

  kosolver.solve();

  return 0;
}

