#include <iostream>
#include <strstream>
#include <vector>

#include "ko/GeometryMesher.hpp"
#include "ko/IntersectionDetector.hpp"

using namespace std;

namespace ko
{


bool validIndex(dolfin::Node ****nodeArray,
		int ix, int iy, int iz, int Nx, int Ny, int Nz)
{
  return
    (ix != -1 && ix < Nx &&
     iy != -1 && iy < Ny &&
     iz != -1 && iz < Nz) &&
    (nodeArray[ix][iy][iz] != 0);
}

void createCell(dolfin::Mesh &m, dolfin::Node ****nodeArray,
		int n1x, int n1y, int n1z,
		int n2x, int n2y, int n2z,
		int n3x, int n3y, int n3z,
		int n4x, int n4y, int n4z,
		int Nx, int Ny, int Nz)
{
  cerr << "Creating cell" << endl;

  if(validIndex(nodeArray, n1x, n1y, n1z, Nx, Ny, Nz) &&
     validIndex(nodeArray, n2x, n2y, n2z, Nx, Ny, Nz) &&
     validIndex(nodeArray, n3x, n3y, n3z, Nx, Ny, Nz) &&
     validIndex(nodeArray, n4x, n4y, n4z, Nx, Ny, Nz))
  {
    dolfin::Node &n1 = *(nodeArray[n1x][n1y][n1z]);
    dolfin::Node &n2 = *(nodeArray[n2x][n2y][n2z]);
    dolfin::Node &n3 = *(nodeArray[n3x][n3y][n3z]);
    dolfin::Node &n4 = *(nodeArray[n4x][n4y][n4z]);
	  
    dolfin::Cell &c = m.createCell(n1, n2, n3, n4);

    cerr << "Created cell: " <<
      n1.id() << " " <<
      n2.id() << " " <<
      n3.id() << " " <<
      n4.id() << " " <<
      endl;
  }
}



GeometryMesher::GeometryMesher()
{
}

void GeometryMesher::init(GtsSurface *surface)
{
  bbTree = gts_bb_tree_surface(surface);

  probePoint = gts_point_new(gts_point_class(), 0, 0, 0);
}

void GeometryMesher::meshgen(double density, bool isCenter)
{
  //double radius = 1 / density / 2 * 1.2;

  double cellSize = 1 / density;

  double radius = cellSize / 2;

  cerr << "cellSize: " << cellSize << endl;

  GtsBBox *largestBox = (GtsBBox *)bbTree->data;

  Vector min(3), max(3), center(3), span(3);

  min(0) = largestBox->x1;
  min(1) = largestBox->y1;
  min(2) = largestBox->z1;

  max(0) = largestBox->x2;
  max(1) = largestBox->y2;
  max(2) = largestBox->z2;

  Vector perturbation(3);

  perturbation(0) = 1e-07;
  perturbation(1) = 1e-07;
  perturbation(2) = 1e-07;

  max.axpy(1, perturbation);
  min.axpy(-1, perturbation);

  center = min;
  center.axpy(1, max);
  center *= 0.5;

  span = max;
  span.axpy(-1, min);


  // some perturbations to make it easier for the geometric predicates

  //mtl::scale(center, 1.01);
  //mtl::scale(span, 0.99);

  int xcount = (int)floor(span(0) * density);
  int ycount = (int)floor(span(1) * density);
  int zcount = (int)floor(span(2) * density);

  cerr << "xcount: " << xcount << endl;
  cerr << "ycount: " << ycount << endl;
  cerr << "zcount: " << zcount << endl;

  dolfin::Node ****nodeArray;
  nodeArray = new dolfin::Node***[xcount];

  for(int ix = 0; ix < xcount; ix++)
  {
    nodeArray[ix] = new dolfin::Node**[ycount];
    for(int iy = 0; iy < ycount; iy++)
    {
      nodeArray[ix][iy] = new dolfin::Node*[zcount];
    }
  }


  for(int ix = 0; ix < xcount; ix++)
  {
    for(int iy = 0; iy < ycount; iy++)
    {
      for(int iz = 0; iz < zcount; iz++)
      {
	nodeArray[ix][iy][iz] = 0;
      }
    }
  }


  //int xcount = (int)rint(span(0) * density);
  //int ycount = (int)rint(span(1) * density);
  //int zcount = (int)rint(span(2) * density);

  //Material *material = new Material;

  //material->name = "default";
  //material->strength = 1.0;

  //physical.materials.push_back(material);

  int counter = 0;

  for(int ix = 0; ix < xcount; ix++)
  {
    for(int iy = 0; iy < ycount; iy++)
    {
      for(int iz = 0; iz < zcount; iz++)
      {
	Vector p(3);

	if(isCenter)
	{
	  p(0) = center(0) - span(0) / 2 + cellSize * ix;
	  p(1) = center(1) - span(1) / 2 + cellSize * iy;
	  p(2) = center(2) - span(2) / 2 + cellSize * iz;
	}
	else
	{
	  p(0) = min(0) + radius + cellSize * ix;
	  p(1) = min(1) + radius + cellSize * iy;
	  p(2) = min(2) + radius + cellSize * iz;
	}

	///*
	// We want to make sure we're filling up all the way to the boundary.
	// We sample at half the cell size, and if any of the samples are
	// positive we fill the cell.

	double offset = cellSize / 4;
	
	int refineCounter = 0;
	
	for(int i = -1; i < 2; i += 2)
	{
	  for(int j = -1; j < 2; j += 2)
	  {
	    for(int k = -1; k < 2; k += 2)
	    {
	      Vector p2(3);
	      
	      p2 = p;
	      
	      p2(0) += i * offset;
	      p2(1) += j * offset;
	      p2(2) += k * offset;
	      
	      if(inside(p2))
	      {
		refineCounter++;
	      }
	    }
	  }
	}
	
	cerr << refineCounter << endl;

	if(refineCounter > 0)
	{
	  dolfin::Node &n = mesh.createNode(0, 0, 0);

	  n.coord().x = p(0);
	  n.coord().y = p(1);
	  n.coord().z = p(2);

	  cerr << "Created node: " <<
	    n.coord().x << " " <<
	    n.coord().y << " " <<
	    n.coord().z <<
	    endl;
	  
	  nodeArray[ix][iy][iz] = &n;

	  counter++;

	}
	//*/
      }
    }
  }




  for(int ix1 = 0; ix1 < xcount; ix1++)
  {
    for(int iy1 = 0; iy1 < ycount; iy1++)
    {
      for(int iz1 = 0; iz1 < zcount; iz1++)
      {
	int pos = (ix1 % 2) + (iy1 % 2) + (iz1 % 2);
	
	std::cerr <<
	  "pos: " << pos << " " <<
	  "ix1: " << ix1 << " " <<
	  "iy1: " << iy1 << " " <<
	  "iz1: " << iz1 <<
	  endl;
	
	if((pos % 2 == 0))
	{
	  // We generate a diamond shape of tetrahedrons around (ix1, iy1, iz1)
	  
	  // Top of pyramid
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 - 1,
		     ix1 + 0, iy1 + 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 - 1,
		     ix1 - 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 - 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 + 1,
		     ix1 + 0, iy1 + 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 + 1,
		     ix1 + 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  // Bottom of pyramid
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 - 1,
		     ix1 + 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 - 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 - 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 - 1,
		     ix1 + 0, iy1 - 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 + 1,
		     ix1 - 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 - 1, iz1 + 0,
		     xcount, ycount, zcount);
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 0, iz1 + 1,
		     ix1 + 0, iy1 - 1, iz1 + 0,
		     xcount, ycount, zcount);
	}
	
	if(pos == 3)
	{
	  ///*
	  // We fill in the holes between the diamonds around (ix1, iy1, iz1)
	  
	  // Upper part (rotate ccw around y axis)
	  
	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 + 1,
		     ix1 + 1, iy1 + 1, iz1 + 0,
		     ix1 + 0, iy1 + 1, iz1 + 1,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 + 1, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 - 1,
		     ix1 + 0, iy1 + 1, iz1 - 1,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 1, iz1 - 1,
		     ix1 - 1, iy1 + 0, iz1 - 1,
		     ix1 - 1, iy1 + 1, iz1 + 0,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 0, iy1 + 1, iz1 + 1,
		     ix1 - 1, iy1 + 1, iz1 + 0,
		     ix1 - 1, iy1 + 0, iz1 + 1,
		     xcount, ycount, zcount);

	  // Lower part (rotate ccw around y axis)

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 - 1, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 + 1,
		     ix1 + 0, iy1 - 1, iz1 + 1,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 + 1, iy1 + 0, iz1 - 1,
		     ix1 + 1, iy1 - 1, iz1 + 0,
		     ix1 + 0, iy1 - 1, iz1 - 1,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 - 1, iy1 + 0, iz1 - 1,
		     ix1 + 0, iy1 - 1, iz1 - 1,
		     ix1 - 1, iy1 - 1, iz1 + 0,
		     xcount, ycount, zcount);

	  createCell(mesh, nodeArray,
		     ix1 + 0, iy1 + 0, iz1 + 0,
		     ix1 - 1, iy1 - 1, iz1 + 0,
		     ix1 + 0, iy1 - 1, iz1 + 1,
		     ix1 - 1, iy1 + 0, iz1 + 1,
		     xcount, ycount, zcount);
	  //*/
	}
      }
    }
  }

  mesh.init();

  // Remove unconnected nodes

  std::vector<dolfin::Node *> nodes;

  for (dolfin::NodeIterator n(&mesh); !n.end(); ++n)
  {
    if((*n).noCellNeighbors() == 0)
    {
      cerr << "Marking node for removal" << endl;

      nodes.push_back(&(*n));
    }
  }

  for(std::vector<dolfin::Node *>::iterator it = nodes.begin();
      it != nodes.end(); it++)
  {
    mesh.remove(*(*it));
  }

  mesh.init();

  cerr << "finished meshing" << endl;


}

bool GeometryMesher::inside(Vector &p)
{
  gts_point_set(probePoint, p(0), p(1), p(2));

  /*
  GSList *boxlist = 0;

  boxlist = gts_bb_tree_stabbed(bbTree, probePoint);

  for(GSList *iter = boxlist; iter != 0; g_slist_next(iter))
  {
    GtsBBox *box = (GtsBBox *)iter->data;
    GtsObject *obj = (GtsObject *)box->bounded;

    if(gts_object_is_from_class(obj, gts_triangle_class()))
    {
      GtsTriangle *triangle = (GtsTriangle *)obj;

      

    }
  }
  */

  return gts_point_is_inside_surface(probePoint, bbTree, false);
}



}
